﻿<#
    .SYNOPSIS
    Returns current connection count for all front end servers in a given pool
    including a breakdown of connection by client, frontend server and users.

    It can also be used to return connection information on an individual user.

    .DESCRIPTION
    This program will return a connection count for a given pool. The program
    can be edited to set a default pool. You will also be able to get
    information on an individual user by providing the users SamAccountName.
    As well as listing all the connected users to the default pool.

    NOTE: In order to gain remote access to each front end server's
    RTCLOCAL database where connection information is found,
    you need to open two local Windows firewall ports. Please see the installation
    info on the blog post for details.

    .NOTES
    Version               : 3.4 - See changelog at https://ucunleashed.com/681 for fixes & changes introduced with each version
    Wish list						  : Better error trapping
    Rights Required			  : TBD
    Sched Task Required	  : No
    Lync/Skype4B Version  : Lync Server 2013, Skype for Business Server 2015
    Author(s)    				  : Pat Richard (pat@innervation.com) 	https://ucunleashed.com @patrichard
                          : Tracy A. Cerise (tracy@uky.edu)
                          : Mahmoud Badran (v-mabadr@microsoft.com)
    Donations             : https://www.paypal.me/PatRichard  										
    Dedicated Post			  : https://ucunleashed.com/269
    Disclaimer            : You running this script/function means you will not blame the author(s) if this breaks your stuff. This
                            script/function is provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties
                            including, without limitation, any implied warranties of merchantability or of fitness for a particular
                            purpose. The entire risk arising out of the use or performance of the sample scripts and documentation
                            remains with you. In no event shall author(s) be held liable for any damages whatsoever (including, 
                            without limitation, damages for loss of business profits, business interruption, loss of business
                            information, or other pecuniary loss) arising out of the use of or inability to use the script or
                            documentation. Neither this script/function, nor any part of it other than those parts that are explicitly
                            copied from others, may be republished without author(s) express written permission. Author(s) retain the
                            right to alter this disclaimer at any time. For the most up to date version of the disclaimer, see 
                            https://ucunleashed.com/code-disclaimer.
    Acknowledgements 		  : http://blogs.technet.com/b/meacoex/archive/2011/07/19/list-connections-and-users-connected-to-lync-registrar-pool.aspx
                            This program's database connection information was originally taken from the "List Connections to Registrar Pools" submitted by Scott Stubberfield and Nick Smith from Microsoft to the Lync 2010 PowerShell blog  (http://blogs.technet.com/b/csps/) on June 10, 2010.
    Assumptions					  : ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
    Limitations					  :
    Known issues				  : None yet, but I'm sure you'll find some!

    .LINK
    https://ucunleashed.com/269

    .EXAMPLE
    .\Get-CsConnections.ps1

    Description
    -----------
    Returns information on all connections on all frontend servers in the
    specified pool or server.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN]

    Description
    -----------
    Returns information on all connections on all frontend servers in the
    Lync Server 2010 pool FQDN given with the pool parameter.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -Is2013

    Description
    -----------
    Returns information on all connections on all frontend servers in the
    Lync Server 2013 pool FQDN given with the pool parameter.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool name]

    Description
    -----------
    Returns information on all connections on all Lync Server frontend servers
    in the pool NetBIOS name given with the pool parameter. The script will append
    the %userdnsdomain% to build the FQDN.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -SIPAddress userid@sip.domain

    Description
    -----------
    Returns all connection information for the given user including which
    frontend server connected to, how many connections and which clients
    connected with. If the environment contains only one SIP domain, only the prefix
    is required.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -FilePath c:\path\to\file\filename.csv

    Description
    -----------
    Returns information on all connections on all frontend servers in the
    pool and in addition writes out all the raw connection information
    into the filename specified.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -IncludeUsers

    Description
    -----------
    Returns information on all connections on all frontend servers in the
    specified pool and in addition writes out all the users connection
    information. This does NOT include system related connections.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -IncludeUsers -UserHighConnectionFlag 5

    Description
    -----------
    Returns information on all connections on all frontend servers in the
    specified pool and in addition writes out all the users connection
    information. This does NOT include system related connections. The
    UserHighConnectionFlag integer will color	all users with that many
    (or more) connections in red. The default is 4.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -IncludeHighUsers -UserHighConnectionFlag 5

    Description
    -----------
    Returns information on all connections on all frontend servers in the
    specified pool and in addition writes out all the users connection
    information for users with high number of connections. This does NOT
    include system related connections. The UserHighConnectionFlag integer
    will color	all users with that many (or more) connections in red.
    The default is 4.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -IncludeUsers -IncludeSystem

    Description
    -----------
    Returns information on all connections on all frontend servers in the
    specified pool and in addition writes out all the users connection
    information. This includes any system	related connections.

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -ClientVersion [version number]

    Description
    -----------
    Returns all connection information for only clients that contain the specified client version.
    Also lists the users connecting with that version (same as -IncludeUsers option).

    .EXAMPLE
    .\Get-CsConnections.ps1 -PoolFqdn [pool FQDN] -ShowFullClient

    Description
    -----------
    Returns extended information for clients including mobile device OS and type.

    .INPUTS
    None. You cannot pipe objects to this script.
#>
#Requires -Version 3.0

[CmdletBinding(SupportsShouldProcess, SupportsPaging, DefaultParameterSetName = 'Default', HelpUri = 'https://ucunleashed.com/269')]
param(
  # Defines the front end pool to query for information. Cannot be used with -Server or -SIPAddress. If not specified and there is only a single front end pool, the script will automatically determine the FQDN.
  [parameter(ValueFromPipelineByPropertyName)]
  [string] $PoolFqdn,

  # Defines a specific server to query for information. Cannot be used with -PoolFqdn or -SIPAddress.
  [parameter(ValueFromPipelineByPropertyName)]
  [string] $Server,

  # Defines the SIP address for a specific user to query. Cannot be used with -PoolFqdn or -Server.
  [parameter(ValueFromPipelineByPropertyName)]
  [string] $SipAddress,

  # Defines the file path for the exported .csv file.
  [parameter(ValueFromPipelineByPropertyName)]
  [string] $FilePath,

  # Displays connected users and their connection counts. Users with a number of connections that meets or exceeds the -UserHighConnectionFlag count are colored in red. Cannot be used with -IncludeHighUsers.
  [parameter(ValueFromPipelineByPropertyName)]
  [switch] $IncludeUsers,

  # Displays connected users who meet or exceed the -UserHighConnectionFlag value. Users who are below the value are in white. Users who exceed the value are in red. Cannot be used with -IncludeUsers.
  [parameter(ValueFromPipelineByPropertyName)]
  [switch] $IncludeHighUsers,

  # Includes accounts used by the system (non-users).
  [parameter(ValueFromPipelineByPropertyName)]
  [switch] $IncludeSystem,

  # Defines the value at which users are considered to have a high number of connections. Defaults to 4. Cannot exceed the number configured for MaxEndpointsPerUser in the global RegistrarConfiguration.
  [parameter(ValueFromPipelineByPropertyName)]
  [ValidateRange(2,10)]
  [int] $UserHighConnectionFlag = 4,

  # Filter results to a specific version of the client. Helpful with -IncludeUsers to determine who's using a specific version of the client.
  [parameter(ValueFromPipelineByPropertyName)]
  [string] $ClientVersion,

  # Displays full client version information. Useful for viewing longer mobile device info.
  [parameter(ValueFromPipelineByPropertyName)]
  [switch] $ShowFullClient,

  # Queries Lync environment for total number of enabled users, and also determines percentage of total users that are currently connected.
  [parameter(ValueFromPipelineByPropertyName)]
  [switch] $ShowTotal,

  # Specifies that the server the script is connecting to is running Lync Server 2013. This parameter is not required when connecting to a pool, since the script will auto detect the version.
  [parameter(ValueFromPipelineByPropertyName)]
  [switch] $Is2013,

  # Skips the check for script age
  [parameter(ValueFromPipelineByPropertyName)]
  [switch] $SkipUpdateCheck,
	
  # Defines the location for any downloaded files. Defaults to "c:\_install". Additionally, log files generated by this script are located in a subfolder of TargetFolder called "logs". TargetFolder does not support paths with spaces, but does support non-hidden UNC paths.
  [string] $TargetFolder = "$env:SystemDrive\_Install"
)
[version] $ScriptVersion = '3.4'
[int] $Article = 269
[string] $LogPath = "$TargetFolder\logs\$env:ComputerName" + ' {0:yyyy-MM-dd hh-mmtt}.log' -f (Get-Date)
[string] $LogDivider = '------------------------------'
[string] $ConsoleDivider = '------------------------------------------------------------------------------'
[string] $ConsoleDividerShort = '--------------------------------------------------------'

#region functions

function Get-Data {
  [CmdletBinding(SupportsShouldProcess)]
  param (
    [parameter(ValueFromPipelineByPropertyName)]
    [string] $SipAddress,

    [parameter(ValueFromPipelineByPropertyName)]
    [string] $Server,

    [parameter(ValueFromPipelineByPropertyName)]
    [string] $ClientVersion
  )

  ##############################################################################################
  # Went to using a named parameter for this function due to the
  # way Powershell does its thing with parameter passing, which
  # is NOT GOOD!  At any rate, need to call this function
  # as you would from a command line: Get-Data -sipAddr "value"
  # -server "value"
  #
  # Also, assuming a value of NULL for the SIP address of an
  # individual user, mostly to use this for finding overall
  # values, only occasionally to seek specific users.
  ##############################################################################################

  if ($SipAddress) {
    [string] $whereClause = "where R.UserAtHost = '$SipAddress' "
  } else {
    if ($IncludeSystem){
      [string] $whereClause = $null
    }elseif ($ClientVersion){
      [string] $whereClause = "where upper(RE.ClientApp) like upper('%$ClientVersion%') and R.UserAtHost not like 'RtcApplication-%' "
    }else{
      [string] $whereClause = "where R.UserAtHost not like 'RtcApplication-%' "
    }
  }

  #Define SQL Connection String
  [string] $connstring = "server=$server\rtclocal;database=rtcdyn;trusted_connection=true;"

  #Define SQL Command
  [object] $command = New-Object -TypeName System.Data.SqlClient.SqlCommand

  if (($PoolFqdn -and (Get-CsService -PoolFqdn $PoolFqdn -Registrar).Version -ge 6) -or $Is2013){
    # SQL query for Lync Server 2013
    $command.CommandText = "Select (cast (RE.ClientApp as varchar (100))) as ClientVersion, R.UserAtHost as UserName, RA.Fqdn `
      From rtcdyn.dbo.RegistrarEndpoint RE `
      Inner Join rtcdyn.dbo.Endpoint EP on RE.EndpointId = EP.EndpointId `
      Inner Join rtc.dbo.Resource R on R.ResourceId = RE.OwnerId `
      Inner Join rtcdyn.dbo.Registrar RA on EP.RegistrarId = RA.RegistrarId `
      $whereClause `
    Order By ClientVersion, UserName"
  }else{
    # SQL query for Lync Server 2010
    $command.CommandText = "Select (cast (RE.ClientApp as varchar (100))) as ClientVersion, R.UserAtHost as UserName, FE.Fqdn `
      From rtcdyn.dbo.RegistrarEndpoint RE `
      Inner Join rtcdyn.dbo.Endpoint EP on RE.EndpointId = EP.EndpointId `
      Inner Join rtc.dbo.Resource R on R.ResourceId = RE.OwnerId `
      Inner Join rtcdyn.dbo.FrontEnd FE on EP.RegistrarId = FE.FrontEndId `
      $whereClause `
    Order By ClientVersion, UserName"
  }

  [object] $connection = New-Object -TypeName System.Data.SqlClient.SqlConnection
  $connection.ConnectionString = $connstring
  $connection.Open()

  $command.Connection = $connection

  [object] $sqladapter = New-Object -TypeName System.Data.SqlClient.SqlDataAdapter
  $sqladapter.SelectCommand = $command

  [object] $results = New-Object -TypeName System.Data.Dataset
  $recordcount = $sqladapter.Fill($results)
  $connection.Close()
  return $Results.Tables[0]
} # end function Get-Data

function Get-UpdateInfo {
  <#
      .SYNOPSIS
      Queries an online XML source for version information to determine if a new version of the script is available.

      .DESCRIPTION
      Queries an online XML source for version information to determine if a new version of the script is available.

      .NOTES
      Version               : 1.5 - See changelog at https://ucunleashed.com/3168 for fixes & changes introduced with each version
      Wish list             : Better error trapping
      Rights Required       : N/A
      Sched Task Required   : No
      Lync/Skype4B Version  : N/A
      Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
      Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
      Donations             : https://www.paypal.me/PatRichard
      Dedicated Post        : https://ucunleashed.com/3168
      Disclaimer            : You running this script/function means you will not blame the author(s) if this breaks your stuff. This script/function 
                              is provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including, without limitation, 
                              any implied warranties of merchantability or of fitness for a particular purpose. The entire risk arising out of the use 
                              or performance of the sample scripts and documentation remains with you. In no event shall author(s) be held liable for 
                              any damages whatsoever (including, without limitation, damages for loss of business profits, business interruption, loss 
                              of business information, or other pecuniary loss) arising out of the use of or inability to use the script or 
                              documentation. Neither this script/function, nor any part of it other than those parts that are explicitly copied from 
                              others, may be republished without author(s) express written permission. Author(s) retain the right to alter this 
                              disclaimer at any time. For the most up to date version of the disclaimer, see https://ucunleashed.com/code-disclaimer.
      Acknowledgements      : Reading XML files 
                            http://stackoverflow.com/questions/18509358/how-to-read-xml-in-powershell
                            http://stackoverflow.com/questions/20433932/determine-xml-node-exists
                          : TLS 1.2 fixup (via @GreigInSydney)
                            https://github.com/chocolatey/choco/wiki/Installation#installing-with-restricted-tls
      Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
      Limitations           : 
      Known issues          : 

      .EXAMPLE
      Get-UpdateInfo -Article 123

      Description
      -----------
      Runs function to check for updates to article/script 123.

      .INPUTS
      None. You cannot pipe objects to this script.
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpUri = 'https://ucunleashed.com/3168')]
  param (
    # Specifies the blog article number. This is used to check the XML file for version info.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [int] $Article,
    
    # Path to the XML file that contains the update information.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [string] $XmlPath = 'https://ucunleashed.com/downloads/version.xml'
  )
  [bool] $HasInternetAccess = ([Activator]::CreateInstance([Type]::GetTypeFromCLSID([Guid]'{DCB00C01-570F-4A9B-8D69-199FDBA5723B}')).IsConnectedToInternet)
  if ($HasInternetAccess){
    Write-Log -Level Info -Message 'Performing update check'
    # ------------------ TLS 1.2 fixup
    $SecurityProtocolSettingsOriginal = [Net.ServicePointManager]::SecurityProtocol
    Write-Log -Level Info -Message "Original security protocol settings: $SecurityProtocolSettingsOriginal" -NoConsole
    try {
      # Set TLS 1.2 (3072). Use integers because the enumeration values for TLS 1.2 won't exist in .NET 4.0, even though they are 
      # addressable if .NET 4.5+ is installed (.NET 4.5 is an in-place upgrade).
      Write-Log -Level Info -Message 'New security protocol settings: Tls12 (3072)' -NoConsole -Indent 1
      [Net.ServicePointManager]::SecurityProtocol = 3072
      # Need to error check here to ensure it was actually changed.
    } catch {
      Write-Log -Level Warn -Message 'Unable to set PowerShell to use TLS 1.2 due to old .NET Framework installed.' -Indent 1
    }
    # ------------------ end TLS 1.2 fixup
    Write-Log -Level Info -Message "Retrieving update info for article $Article from $XmlPath" -NoConsole -Indent 1
    [xml] $xml = (New-Object -TypeName System.Net.WebClient).DownloadString($XmlPath)
    # Reinstate original SecurityProtocol settings from before the TLS 1.2 fixup
    Write-Log -Level Info -Message "Restoring original security protocol settings to $SecurityProtocolSettingsOriginal" -NoConsole -Indent 1
    [Net.ServicePointManager]::SecurityProtocol = $SecurityProtocolSettingsOriginal    
    try{
      [object] $ArticleXml = Select-Xml -Xml $xml -XPath "//article[@id='$($article)']"
      [string] $Ga = $ArticleXml.Node.Version.Trim()
      Write-Log -Level Info -Message "GA version according to XML is $Ga" -NoConsole -Indent 1
    
      [string] $Title = $ArticleXml.Node.Title.Trim()
      if (-Not ([string] $DownloadUrl = $ArticleXml.Node.DownloadUrl.Trim())){
        [string] $DownloadUrl = "https://ucunleashed.com/$article#downloads"
      }
      [string] $changelog = 'This version includes: ' + $ArticleXml.Node.Changelog.Trim()

      if ($Ga -gt $ScriptVersion){
        Write-Log -Level Warn -Message "Outdated version. Version $Ga is latest version. Prompting user to upgrade." -NoConsole
        if ((New-Popup -Message "Version $Ga is now available. $changelog `n`nWould you like to download it?" -Title "New version of $Title available!" -Buttons YesNo -Icon Question) -eq 6){
          Write-Log -Level Info -Message "User elected to download version $Ga." -NoConsole -Indent 1
          Write-Log -Level Info -Message "Opening Internet Explorer and going to $DownloadUrl" -NoConsole -Indent 1
          Start-Process -FilePath $DownloadUrl
          Write-Log -Level Warn -Message "Script is exiting. Please start the new version of the script once you've downloaded it." -NoLog
          Stop-Script
        }else{
          Write-Log -Level Warn -Message "User elected NOT to download version $Ga." -Indent 1 -NoConsole
        }
      }elseif ($Ga -eq $ScriptVersion){
        Write-Log -Level Info -Message "Current version. Version $Ga is latest version." -NoConsole
          }elseif (-Not ($Ga)) {
        Write-Log -Level Warn -Message 'No version info available' -NoConsole
      }
    }
    catch{}
  }else{
    Write-Log -Level Warn -Message 'Unable to check online for update info.' -NoConsole
  }
  Write-Log -Level Info -Message 'Completed update check' -NoConsole -Indent 1
} # end function Get-UpdateInfo

function New-Popup {
  <#
      .Synopsis
      Display a Popup Message

      .Description
      This command uses the Wscript.Shell PopUp method to display a graphical message
      box. You can customize its appearance of icons and buttons. By default the user
      must click a button to dismiss but you can set a timeout value in seconds to
      automatically dismiss the popup.

      The command will write the return value of the clicked button to the pipeline:
      OK     = 1
      Cancel = 2
      Abort  = 3
      Retry  = 4
      Ignore = 5
      Yes    = 6
      No     = 7

      If no button is clicked, the return value is -1.

      .Example
      PS C:\> new-popup -message "The update script has completed" -title "Finished" -time 5

      This will display a popup message using the default OK button and default
      Information icon. The popup will automatically dismiss after 5 seconds.

      .Notes
      Last Updated: April 8, 2013
      Version     : 1.0

      .Inputs
      None

      .Outputs
      integer

      Null   = -1
      OK     = 1
      Cancel = 2
      Abort  = 3
      Retry  = 4
      Ignore = 5
      Yes    = 6
      No     = 7
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging)]
  Param (
    # Message to be displayed in the popup message dialog box.
    [Parameter(Position = 0, Mandatory, HelpMessage = 'Enter a message for the popup')]
    [ValidateNotNullorEmpty()]
    [string] $Message,

    # Title of the popup dialog box.
    [Parameter(Position = 1, Mandatory, HelpMessage = 'Enter a title for the popup')]
    [ValidateNotNullorEmpty()]
    [string] $Title,

    # If defined, automatically dismisses the dialog box after the specified number of seconds. Defaults to 0, which requires a mouse click.
    [Parameter(Position = 2)]
    [ValidateScript({$_ -ge 0})]
    [int] $Time = 0,

    # Type of buttons to display.
    [Parameter(Position = 3)]
    [ValidateNotNullorEmpty()]
    [ValidateSet('OK','OKCancel','AbortRetryIgnore','YesNo','YesNoCancel','RetryCancel')]
    [string] $Buttons = 'OK',

    # Icon set to use in dialog box.
    [Parameter(Position = 4)]
    [ValidateNotNullorEmpty()]
    [ValidateSet('Stop','Question','Exclamation','Information' )]
    [string] $Icon = 'Information'
  ) # end of param block

  #convert buttons to their integer equivalents
  Switch ($Buttons) {
    'OK'               {$ButtonValue = 0}
    'OKCancel'         {$ButtonValue = 1}
    'AbortRetryIgnore' {$ButtonValue = 2}
    'YesNo'            {$ButtonValue = 4}
    'YesNoCancel'      {$ButtonValue = 3}
    'RetryCancel'      {$ButtonValue = 5}
  }

  #set an integer value for Icon type
  Switch ($Icon) {
    'Stop'        {$iconValue = 16}
    'Question'    {$iconValue = 32}
    'Exclamation' {$iconValue = 48}
    'Information' {$iconValue = 64}
  }

  #create the COM Object
  Try {
    $wshell = New-Object -ComObject Wscript.Shell -ErrorAction Stop
    #Button and icon type values are added together to create an integer value
    $wshell.Popup($Message,$Time,$Title,$ButtonValue+$iconValue)
  }

  Catch {
    #You should never really run into an exception in normal usage
    Write-Log -Level Warn -Message 'Failed to create Wscript.Shell COM object'
    Write-Log -Level Warn -Message $_.exception.message
  }
} # end function New-Popup

function Set-ModuleStatus {
  <#
      .SYNOPSIS
      Imports a specified PowerShell module, with error checking.

      .DESCRIPTION
      Imports a specified PowerShell module, with error checking.

      .NOTES
      Version               : 1.6
      Wish list             : Better error trapping
      Rights Required       : None
      Sched Task Required   : No
      Lync/Skype4B Version  : N/A
      Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
      Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
      Donations             : https://www.paypal.me/PatRichard
      Dedicated Post        :
      Disclaimer            : You running this script/function means you will not blame the author(s) if this breaks your stuff. This script/function 
                              is provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including, without limitation, 
                              any implied warranties of merchantability or of fitness for a particular purpose. The entire risk arising out of the use 
                              or performance of the sample scripts and documentation remains with you. In no event shall author(s) be held liable for 
                              any damages whatsoever (including, without limitation, damages for loss of business profits, business interruption, loss 
                              of business information, or other pecuniary loss) arising out of the use of or inability to use the script or 
                              documentation. Neither this script/function, nor any part of it other than those parts that are explicitly copied from 
                              others, may be republished without author(s) express written permission. Author(s) retain the right to alter this 
                              disclaimer at any time. For the most up to date version of the disclaimer, see https://ucunleashed.com/code-disclaimer.
      Acknowledgements      :
      Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
      Limitations           :
      Known issues          : None yet, but I'm sure you'll find some!

      .LINK

      .EXAMPLE
      .\

      Description
      -----------


      .INPUTS
      None. You cannot pipe objects to this script.
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpUri = 'https://ucunleashed.com/938')]
  param  (
    [parameter(ValueFromPipeline, ValueFromPipelineByPropertyName, Mandatory, HelpMessage = 'No module name specified!')]
    [ValidateNotNullOrEmpty()]
    [string] $Name
  ) # end of param block
  PROCESS{
    # Executes once for each pipeline object
    # the $_ variable represents the current input object
    if (-Not (Get-Module -name "$name")) {
      if (Get-Module -ListAvailable | Where-Object Name -eq "$name") {
        Import-Module -Name "$name"
        # module was imported
        # return $true
      } 
    }
  } # end PROCESS
} # end function Set-ModuleStatus

function Remove-ScriptVariable {
	<#
  .SYNOPSIS
    Removed variables defined by a script.

  .DESCRIPTION
    Removed variables defined by a script. Beneficial if the PowerShell session is going to remain open, as removing the variables frees up the memory used.

  .NOTES
    Version               : 1.2
    Wish list             : Better error trapping
    Rights Required       : N/A
    Sched Task Required   : No
    Lync/Skype4B Version  : N/A
    Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
    Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
    Donations             : https://www.paypal.me/PatRichard
    Dedicated Post        :	https://www.ucunleashed.com/247
    Disclaimer            : You running this script means you won't blame author(s) if this breaks your stuff. This script is
                            provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including,
                            without limitation, any implied warranties of merchantability or of fitness for a particular
                            purpose. The entire risk arising out of the use or performance of the sample scripts and
                            documentation remains with you. In no event shall author(s) be liable for any damages whatsoever
                            (including, without limitation, damages for loss of business profits, business interruption,
                            loss of business information, or other pecuniary loss) arising out of the use of or inability
                            to use the script or documentation. Neither this script, nor any part of it other than those
                            parts that are explicitly copied from others, may be republished without author(s) express written 
                            permission.
    Acknowledgements      :
    Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
    Limitations           :
    Known issues          : None yet, but I'm sure you'll find some!

  .LINK
		https://www.ucunleashed.com/247

  .EXAMPLE
    Remove-ScriptVariables

    Description
    -----------
    Removes the variables in memory that were created when a .ps1 script is run.

  .INPUTS
    You can pipe file path objects to this script.
  #>
	[cmdletBinding(SupportsShouldProcess)]
	param(
		[Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
		[ValidateNotNullOrEmpty()]
		[string] $path
	)
	process{
		Write-Verbose -Message 'Begin removing script variables from memory'
		$result = Get-Content -Path $path |  
		ForEach-Object { 
			if ($_ -match '(\$.*?)\s*=') {      
				$matches[1]  | Where-Object { $_ -notlike '*.*' -and $_ -notmatch 'result' -and $_ -notmatch 'env:'}  
			}  
		}  
		ForEach ($v in ($result | Sort-Object | Get-Unique)){		
			Remove-Variable -Name ($v.replace('$','')) -ErrorAction SilentlyContinue
		}
	}
	end{
		Write-Verbose -Message 'Completed removing script variables from memory'
	}
} # end function Remove-ScriptVariable

function Test-IsSigned {
  <#
      .SYNOPSIS

      .DESCRIPTION

      .NOTES
      Version							: 1.0 - See changelog at
      Wish list						: Better error trapping
      Rights Required			: Local administrator on server
      Sched Task Required	: No
      Lync Server Version	: N/A
      Author/Copyright		: © Pat Richard, Lync MVP - All Rights Reserved
      Email/Blog/Twitter	: pat@innervation.com 	https://ucunleashed.com @patrichard
      Dedicated Blog Post	: https://ucunleashed.com/1697
      Disclaimer   				: You running this script means you won't blame me if this breaks your stuff. This script is
                        provided AS IS without warranty of any kind. I disclaim all implied warranties including,
                        without limitation, any implied warranties of merchantability or of fitness for a particular
                        purpose. The entire risk arising out of the use or performance of the sample scripts and
                        documentation remains with you. In no event shall I be liable for any damages whatsoever
                        (including, without limitation, damages for loss of business profits, business interruption,
                        loss of business information, or other pecuniary loss) arising out of the use of or inability
                        to use the script or documentation.
      Assumptions					: ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
      Limitations					:
      Known issues				: None yet, but I'm sure you'll find some!
      Acknowledgements 		:

      .LINK


      .EXAMPLE
      PS C:\>

      Description
      -----------


      .INPUTS
      None. You cannot pipe objects to this function.

      .OUTPUTS
      Boolean output

  #>
  [CmdletBinding(SupportsShouldProcess)]
  param (
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [string] $FilePath = $MyInvocation.ScriptName
  )

  BEGIN{
    # Executes once before first item in pipeline is processed
    # the $_ variable represents the current input object
  } # end BEGIN

  PROCESS{
    # Executes once for each pipeline object
    # the $_ variable represents the current input object
    if ((Get-AuthenticodeSignature -FilePath $FilePath).Status -eq 'Valid'){
      <#
          if ((Get-ExecutionPolicy) -ne "AllSigned"){
          Write-Warning "You could use an ExecutionPolicy of `"AllSigned`" when using this script. AllSigned is a higher level of security. For more information, see http://technet.microsoft.com/en-us/library/ee176961.aspx"
          }
      #>
    }
  } # end PROCESS

  END{
    # Executes once after last pipeline object is processed
    # the $_ variable represents the current input object
  } # end END
} # end function Test-IsSigned

function Write-Log {
  <#
      .SYNOPSIS
      Extensive function to write data to either the console screen, a log file, and/or a Windows event log.

      .DESCRIPTION
      Extensive function to write data to either the console screen, a log file, and/or a Windows event log. Data can be written as info, warning, error, and includes indentation, time stamps, etc.

      .NOTES
      Version               : 3.2
      Wish list             : Better error trapping
      Rights Required       : Local administrator on server if writing to event log(s)
      Sched Task Required   : No
      Lync/Skype4B Version  : N/A
      Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
      Email/Blog/Twitter		: pat@innervation.com 	https://www.ucunleashed.com @patrichard
      Donations             : https://www.paypal.me/PatRichard
      Dedicated Post        : http://poshcode.org/6894
      Disclaimer            : You running this script/function means you will not blame the author(s) if this breaks your stuff. This script/function 
                              is provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including, without limitation, 
                              any implied warranties of merchantability or of fitness for a particular purpose. The entire risk arising out of the use 
                              or performance of the sample scripts and documentation remains with you. In no event shall author(s) be held liable for 
                              any damages whatsoever (including, without limitation, damages for loss of business profits, business interruption, loss 
                              of business information, or other pecuniary loss) arising out of the use of or inability to use the script or 
                              documentation. Neither this script/function, nor any part of it other than those parts that are explicitly copied from 
                              others, may be republished without author(s) express written permission. Author(s) retain the right to alter this 
                              disclaimer at any time. For the most up to date version of the disclaimer, see https://ucunleashed.com/code-disclaimer.
      Acknowledgements      : Based on an original function by Any Arismendi, along with updates by others
                              http://poshcode.org/2566
                            
                              Test for log names and sources
                              http://powershell.com/cs/blogs/tips/archive/2013/06/10/testing-event-log-names-and-sources.aspx
                            
                              Writing to different event logs and sources registered to a single event log
                              http://social.technet.microsoft.com/Forums/en-US/winserverpowershell/thread/e172f039-ce88-4c9f-b19a-0dd6dc568fa0/
      Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
      Limitations           : Writing to event logs requires admin rights
      Known issues          :

      .EXAMPLE
      .\

      Description
      -----------


      .INPUTS
      System.String. You cannot pipe objects to this script.

      .OUTPUTS
      System.String
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpUri = 'http://poshcode.org/6894')]
  param(
    # The type of message to be logged. Alias is 'type'.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateSet('Error', 'Warn', 'Info', 'Verbose')]
    [ValidateNotNullOrEmpty()]
    [string] $Level = 'Info',

    # The message to be logged.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName, Mandatory, HelpMessage = 'No message specified.')]
    [ValidateNotNullOrEmpty()]
    [string] $Message,

    # Specifies that $message should not the sent to the log file.
    [Parameter(ValueFromPipelineByPropertyName)]
    [switch] $NoLog,

    # Specifies to not display the message to the console.
    [Parameter(ValueFromPipelineByPropertyName)]
    [switch] $NoConsole,

    # The number of spaces to indent the message in the log file.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateRange(1,30)]
    [ValidateNotNullOrEmpty()]
    [int] $Indent = 0,

    # Specifies what color the text should be be displayed on the console. Ignored when switch 'NoConsoleOut' is specified.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateSet('Black', 'DarkMagenta', 'DarkRed', 'DarkBlue', 'DarkGreen', 'DarkCyan', 'DarkYellow', 'Red', 'Blue', 'Green', 'Cyan', 'Magenta', 'Yellow', 'DarkGray', 'Gray', 'White')]
    [ValidateNotNullOrEmpty()]
    [String] $ConsoleForeground = 'White',

    # Existing log file is deleted when this is specified. Alias is 'Overwrite'.
    [Parameter(ValueFromPipelineByPropertyName)]
    [Switch] $Clobber,

    # The name of the system event log, e.g. 'Application'. The Skype for Business log is still called 'Lync Server'. Note that writing to the system event log requires elevated permissions.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateSet('Application', 'System', 'Security', 'Lync Server', 'Microsoft Office Web Apps')]
    [ValidateNotNullOrEmpty()]
    [String] $EventLogName,

    # The name to appear as the source attribute for the system event log entry. This is ignored unless 'EventLogName' is specified.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [String] $EventSource = $([IO.FileInfo] $MyInvocation.ScriptName).Name,

    # The ID to appear as the event ID attribute for the system event log entry. This is ignored unless 'EventLogName' is specified.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateRange(1,65535)]
    [ValidateNotNullOrEmpty()]
    [int] $EventID = 1,

    # The text encoding for the log file. Default is ASCII.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateSet('Unicode','Byte','BigEndianUnicode','UTF8','UTF7','UTF32','ASCII','Default','OEM')]
    [ValidateNotNullOrEmpty()]
    [String] $LogEncoding = 'ASCII'
        
    # Destination folder for log files
    #[Parameter(ValueFromPipelineByPropertyName)]
    #[ValidateNotNullOrEmpty()]
    #[string] $LogPath = "$TargetFolder\logs\$env:ComputerName" + " {0:yyyy-MM-dd hh-mmtt}.log" -f (Get-Date),

    # Divider line to be used to separate sections in the log file
    #[Parameter(ValueFromPipelineByPropertyName, ParameterSetName = 'Divider')]
    #[ValidateNotNullOrEmpty()]
    #[string] $LogDivider = '+------------------------------+'
  ) # end of param block
  BEGIN{
    [string]$LogFolder = Split-Path -Path $LogPath -Parent
    if (-not (Test-Path -Path $LogFolder)){
      $null = New-Item -Path $LogFolder -ItemType Directory
    }
  } # end BEGIN
  PROCESS{
    try {
      $Message = $($Message.trim())
      $msg = '{0} : {1} : {2}{3}' -f (Get-Date -Format 'yyyy-MM-dd HH:mm:ss'), $Level.ToUpper(), ('  ' * $Indent), $Message
      if (-not ($NoConsole)){
        switch ($Level) {
          'Error' {$Host.UI.WriteErrorLine("$Message")}
          'Warn' {Write-Warning -Message $Message}
          'Info' {Write-Host $Message -ForegroundColor $ConsoleForeground}
          'Verbose' {Write-Verbose -Message $Message}
        }
      }
      if (-not ($NoLog)){
        if ($Clobber) {
          $msg | Out-File -FilePath $LogPath -Encoding $LogEncoding -Force
        } else {
          $msg | Out-File -FilePath $LogPath -Encoding $LogEncoding -Append
        }
      }
      if ($EventLogName) {
        if (-not $EventSource) {
          [string] $EventSource = $([IO.FileInfo] $MyInvocation.ScriptName).Name
        }

        if(-not [Diagnostics.EventLog]::SourceExists($EventSource)) {
          [Diagnostics.EventLog]::CreateEventSource($EventSource, $EventLogName)
        }

        switch ($Level) {
          'Error' {$EntryType = 'Error'}
          'Warn'  {$EntryType = 'Warning'}
          'Info'  {$EntryType = 'Information'}
          'Verbose' {$EntryType = 'Information'}
          Default  {$EntryType = 'Information'}
        }
        Write-EventLog -LogName $EventLogName -Source $EventSource -EventId 1 -EntryType $EntryType -Message $Message
      }
      $msg = ''
    } # end try
    catch {
      Throw "Failed to create log entry in: '$LogPath'. The error was: '$_'."
    } # end catch
  } # end PROCESS
  END{} # end END
} # end function Write-Log
#endregion functions

Write-Log -Message $LogDivider -NoConsole
if (-Not($SkipUpdateCheck)){
  Write-Log -Level Info -Message 'Checking for update'
  Get-UpdateInfo -Article $Article
}else{
  Write-Log -Level Info -Message 'Skipping update check'
}

Set-ModuleStatus -Name Lync
if ($UserHighConnectionFlag){
  # We have to target the global policy or environments with more than one CsRegistrarConfiguration will barf
  $MaxEndPointsPerUser = (Get-CsRegistrarConfiguration -Identity Global).MaxEndPointsPerUser
  if ($UserHighConnectionFlag -gt $MaxEndPointsPerUser){
    Write-Host "MaxEndPointsPerUser in the global configuration is $MaxEndPointsPerUser. Please specify a number for UserHighConnectionFlag that does not exceed $MaxEndPointsPerUser. You specified $UserHighConnectionFlag." -ForegroundColor Red
    exit
  }
}
if ((-Not ($server)) -and (-Not ($PoolFqdn)) -and ((Get-CsService -Registrar | Measure-Object).count -eq 1)){
  Write-Verbose -Message 'Retrieving pool info'
  $PoolFqdn = (Get-CsService -Registrar).PoolFqdn
  Write-Verbose -Message "Pool is now set to $PoolFqdn"
}
if ((-Not ($server)) -and (-Not ($PoolFqdn))){
  $PoolFqdn = Read-Host -Prompt 'Enter front end pool FQDN'
  Write-Verbose -Message "Pool is now set to $PoolFqdn"
}

#################################################################################################
########################################  Main Program  #########################################
#################################################################################################
# Here is where we pull all the front end server(s) from our topology for the designated
# pool and iterate through them to get current connections from all the servers.
#
# There are several possibilities here:
#	 1. Have collection of frontend servers
#	 2. Have a single frontend server or
#	 3. Have no servers
#  4. User specified a server instead of a pool

if ($PoolFqdn -and (-Not ($server))){
  # set FQDN if $PoolFqdn was specified as just the netbios name
  if (-Not($PoolFqdn -match '\.')){
    Write-Verbose -Message "No FQDN supplied. Building it. This may not work if the pool domain is different than this computer's domain"
    $PoolFqdn = $PoolFqdn+'.'+([DirectoryServices.ActiveDirectory.Domain]::GetComputerDomain()).name
  }
  Write-Host "`n`nChecking these pool servers in pool: " -ForegroundColor Cyan -NoNewLine
  Write-Host $PoolFqdn
  $feServers = Get-CsComputer -Pool $PoolFqdn | Sort-Object -Property identity
  ForEach ($feserver in $feservers){
    Write-Host $($feserver.identity) -ForegroundColor Yellow
  }
}elseif ($server -and (-Not ($PoolFqdn))){
  if (-Not($server -match '\.')){
    Write-Verbose -Message "No FQDN supplied. Building it. This may not work if the server domain is different than this computer's domain"
    $server = $server+'.'+([DirectoryServices.ActiveDirectory.Domain]::GetComputerDomain()).name
  }
  $feServers = Get-CsComputer -Identity $server -ErrorAction SilentlyContinue
  Write-Host "`n`nChecking this server: " -ForegroundColor Cyan -NoNewLine
  Write-Host $server
}

# next line added as recommended by Tristan Griffiths
$OverallRecords = @()

if ($feServers.count) {
  # Frontend pool collection, iterate through them
  for ($i = 0; $i -lt $feServers.count; $i++) {
    if ($SipAddress) {
      if ((-Not($SipAddress -match '@'))-and ((Get-CsSipDomain | Measure-Object).count -eq 1)){
				Write-Verbose -Message 'Incomplete SIP address supplied. Building it.'
				$SipAddress = $SipAddress+'@'+(Get-CsSipDomain).identity
				Write-Verbose -Message "SIP address corrected to $SipAddress"
			}
      Write-Verbose -Message "Gathering info from $($feServers[$i].identity)"
      $data = Get-Data -SipAddress $SipAddress -Server $feServers[$i].identity -ClientVersion $ClientVersion
    } else {
      Write-Verbose -Message "Gathering info from $($feServers[$i].identity)"
      $data = Get-Data -Server $feServers[$i].identity -ClientVersion $ClientVersion
    }

    # Since an individual user's connections are all homed on one server,
    # we won't have data coming back from all front-end servers in the
    # case of searching for a single user
    if ($data) {
      $OverallRecords = $OverallRecords + $data
    }
  }
} elseif ($feServers) {
  # Have a standalone server or a FE pool of only one server
  if ($SipAddress) {
    if ((-Not($SipAddress -match '@'))-and ((Get-CsSipDomain | Measure-Object).count -eq 1)){
			Write-Verbose -Message 'Incomplete SIP address supplied. Building it.'
			$SipAddress = $SipAddress+'@'+(Get-CsSipDomain).identity
			Write-Verbose -Message "SIP address corrected to $SipAddress"
		}
    $data = Get-Data -SipAddress $SipAddress -Server $feServers.identity -ClientVersion $ClientVersion
  } else {
    Write-Verbose -Message "Gathering info from $($feServers.identity)"
    $data = Get-Data -Server $feServers.identity -ClientVersion $ClientVersion
  }

  # Make sure we have data to work with...
  if ($data) {
    $OverallRecords = $data
  }
}else{
  Write-Host 'No servers returned!' -ForegroundColor Red
}

# Check to see if we have any data to act on
if (-Not ($OverallRecords)) {
  Write-Host "`r`nNothing returned from query!`r`n" -ForegroundColor Yellow

  # Nothing else to do
  exit
} else {
  $count = 0
  $userHash = @{}
  $clientHash = @{}
  $serverHash = @{}
  $UserList = @{}

  $OverallRecords | ForEach-object {
    # Each record has three components: Connected Client Version, User's SIP
    # address and the frontend server's FQDN. Here, we'll build a hash
    # for each of these components for each record.

    # Build hash of users

    $UserList = $_.UserName

    if (-Not ($userHash.ContainsKey($_.UserName))) {
      $userHash.add($_.UserName, 1)
    } else {
      $userHash.set_item($_.UserName, ($userHash.get_item($_.UserName) + 1))
    }

    # Build hash of servers
    if (-Not ($serverHash.ContainsKey($_.fqdn))) {
      $serverHash.add($_.fqdn, 1)
    } else {
      $serverHash.set_item($_.fqdn, ($serverHash.get_item($_.fqdn) + 1))
    }

    # Build hash of clients
    # Lets get rid of the extraneous verbage from the client version names, if applicable
    # This merely gives a friendlier output, and helps prevent wordwrap
    if ($_.ClientVersion.contains('(') -and (-Not ($ShowFullClient))) {
      # Get rid of extraneous verbage
      $clientName = $_.ClientVersion.substring(0, $_.ClientVersion.IndexOf('('))
    } else {
      # Have a client name with no extraneous verbage or $ShowFullClient switch specified
      $clientName = $_.ClientVersion
      $clientName = $clientName.replace('Microsoft ','')
      $clientName = $clientName.replace('Office ','')
      $clientName = $clientName.replace('AndroidLync','Android')
      $clientName = $clientName.replace('iPadLync','iPad')
      $clientName = $clientName.replace('iPhoneLync','iPhone')
      $clientName = $clientName.replace('WPLync','WP')
    }

    if (-Not ($clientHash.ContainsKey($clientName))) {
      $clientHash.add($clientName, 1)
    } else {
      $clientHash.set_item($ClientName, ($clientHash.get_item($ClientName) + 1))
    }
    $count++
  }
}

#################################################################################################
####################################  Output Query Results  #####################################
#################################################################################################
# If output to file is chosen, then write out the results and a note to that effect
# then exit

if ($FilePath) {
  $OverallRecords | Export-Csv -Path $FilePath
  Write-Host -ForegroundColor green "`r`nQuery Results written to $FilePath`r`n"
  exit
}

#region ClientVersions
if (-Not ($ShowFullClient)){
  Write-Host ("`r`n`r`n{0, -26}{1, -41}{2, 11}" -f 'Client Version', 'Agent', 'Connections') -ForegroundColor Cyan
}else{
  Write-Host ("`r`n`r`n{0, -64}{1, 14}" -f 'Agent', 'Connections') -ForegroundColor cyan
}
Write-Host $ConsoleDivider -ForegroundColor Cyan

ForEach ($key in $clientHash.keys | Sort-Object -Descending) {
  # Break down client version into its two component parts and print
  # them out along with their respective counts in a nice format
  $index = $key.indexof(' ')

  if ($index -eq '-1') {
    # No second part
    $first = $key
    $second = ' '
  } else {
    # Client version/agent has two main parts
    $first = $key.substring(0, $index)
    $second = $key.substring($index + 1)
  }

  $value = $clientHash.$key
  if (-Not ($ShowFullClient)){
    Write-Host ("{0,-26}{1,-45}{2,7}" -f $($first.trim()), $($second.trim()), $value)
  }else{
    Write-Host ("{0,-73}{1,4}" -f $($second.trim()), $value)
  }
}
 Write-Host $ConsoleDivider -ForegroundColor Cyan
# "{0,-41}{1,37}" -f "Client Versions Connected", $clientHash.count
#endregion ClientVersions

#region FrontEnds
# Frontend Server, Connections
Write-Host ("`r`n`r`n{0,-41}{1,15}" -f 'Front End Servers', 'Connections') -ForegroundColor Cyan
Write-Host $ConsoleDividerShort -ForegroundColor Cyan
ForEach ($key in ($serverHash.keys | Sort-Object)) {
  $value = $serverHash.$key
  [string] $Percent = '('+"{0:P2}" -f ($value/$count)+')'
  Write-Host ("{0,-40}{1,6} {2,9}" -f $($key.ToLower()), $value, $Percent)
}
Write-Host $ConsoleDividerShort -ForegroundColor Cyan
Write-Host ("{0,-40}{1,6}" -f 'Total connections', $count)
#endregion FrontEnds

#region UniqueUsers
# Unique Users, Unique Clients
$UniqueUsers = $userHash.count
Write-Host ("`r`n`r`n{0,-41}{1,15}" -f 'Total Unique Users/Clients', 'Total') -ForegroundColor Cyan
Write-Host $ConsoleDividerShort -ForegroundColor Cyan
Write-Host ("{0,-41}{1,15}" -f 'Users Connected', $userHash.count)
#endregion UniqueUsers

#region ShowTotal
if ($ShowTotal){
  Write-Host 'Calculating data...' -ForegroundColor Yellow -NoNewline
  [int] $TotalUsers = (Get-CsUser | Measure-Object).count
  [int] $TotalEVUsers = (Get-CsUser -filter {EnterpriseVoiceEnabled -eq $True} | Measure-Object).Count
  [string] $TotalPercent = "{0:P2}" -f ($UniqueUsers/$TotalUsers)
  Write-Host "`b`b`b`b`b`b`b`b`b`b`b`b`b`b`b`b`b`b`b" -NoNewline
  Write-Host ("{0,-41}{1,15}" -f 'Lync Enabled Users (Entire organization)', $TotalUsers)
  Write-Host ("{0,-41}{1,15}" -f 'Voice Enabled Users (Entire organization)', $TotalEVUsers)
  Write-Host ("{0,-41}{1,15}" -f 'Percentage of Enabled Users Connected', $TotalPercent)
}
Write-Host ("{0,-41}{1,15}" -f 'Client Versions Connected', $clientHash.count)
Write-Host $ConsoleDividerShort -ForegroundColor Cyan
#endregion ShowTotal

#region IncludeUsers
# Users, Connections
if ($IncludeUsers){
  Write-Host ("`r`n`r`n{0,-45}{1,-11}" -f 'Connected Users', 'Connections') -ForegroundColor Cyan
  Write-Host $ConsoleDividerShort -ForegroundColor Cyan
  ForEach ($key in $userHash.keys | Sort-Object) {
    $value = $userHash.$key
    if ($value -ge $UserHighConnectionFlag){
      Write-Host ("{0,-45}{1,11}" -f $key, $value) -ForegroundColor Red
    }else{
      Write-Host ("{0,-45}{1,11}" -f $key, $value)
    }
  }
  Write-Host "--------------------------------------------------------`r`n" -ForegroundColor Cyan
}
#endregion IncludeUsers

#region HighUsers
# Users, Connections
if ($IncludeHighUsers){
  Write-Host ("`r`n`r`n{0,-45}{1,-11}" -f 'Connected Users', 'Connections') -ForegroundColor Cyan
  Write-Host $ConsoleDividerShort -ForegroundColor Cyan
  ForEach ($key in $userHash.keys | Sort-Object) {
    $value = $userHash.$key
    if ($value -eq $UserHighConnectionFlag){
      Write-Host ("{0,-45}{1,11}" -f $key, $value)
    }elseif ($value -gt $UserHighConnectionFlag){
      Write-Host ("{0,-45}{1,11}" -f $key, $value) -ForegroundColor Red
    }
  }
  Write-Host "--------------------------------------------------------`r`n" -ForegroundColor Cyan
}
#endregion HighUsers

#region SipAddress
if ($SipAddress){
  Write-Host "`r`n`r`n"
  # we have to get creative here as the output from Get-CsUserPoolInfo changed from Lync 2010 to 2013
  if (((Get-Module -Name Lync).Version).Major -eq 4){
    # Lync Server 2010
    $UserPreferredInfo = Get-CsUserPoolInfo -Identity sip:$SipAddress | Select-Object -ExpandProperty PrimaryPoolMachinesInPreferredOrder | Select-Object -Property MachineId,Fqdn
    Write-Host ("{0,-40}" -f "Preferred Connection Order For $SipAddress") -ForegroundColor Cyan
    Write-Host $ConsoleDividerShort -ForegroundColor Cyan
    ForEach ($server in $UserPreferredInfo){
      $machineid = $server.machineid
      $fqdn = $server.fqdn
      Write-Host ("{0,-40}" -f $fqdn)
    }
  }elseif (((Get-Module -Name Lync).Version).Major -ge 5){
    # Lync Server 2013
    $UserPreferredInfo = Get-CsUserPoolInfo -Identity sip:$SipAddress | Select-Object -ExpandProperty PrimaryPoolMachinesInPreferredOrder
    Write-Host ("{0,-40}" -f "Preferred Connection Order For $SipAddress") -ForegroundColor Cyan
    Write-Host $ConsoleDividerShort -ForegroundColor Cyan
    ForEach ($server in $UserPreferredInfo){
      Write-Host ("{0,-40}" -f $server)
    }
  }
  Write-Host "--------------------------------------------------------`n" -ForegroundColor Cyan
}
#endregion SipAddress
Write-Verbose -Message 'Query complete'

Remove-ScriptVariable -Path ($MyInvocation.MyCommand.Definition)

# SIG # Begin signature block
# MIIcdgYJKoZIhvcNAQcCoIIcZzCCHGMCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQU6Kiy4X1X1CbEclhoLgS6y3Lr
# 1XmgghelMIIFLjCCBBagAwIBAgIQAWzpe+2ITyY5e1fzGr5MLDANBgkqhkiG9w0B
# AQsFADByMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYD
# VQQLExB3d3cuZGlnaWNlcnQuY29tMTEwLwYDVQQDEyhEaWdpQ2VydCBTSEEyIEFz
# c3VyZWQgSUQgQ29kZSBTaWduaW5nIENBMB4XDTE3MTAyNTAwMDAwMFoXDTE4MTAz
# MDEyMDAwMFowazELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAk1JMRkwFwYDVQQHExBT
# dGVybGluZyBIZWlnaHRzMRkwFwYDVQQKExBJbm5lcnZhdGlvbiwgTExDMRkwFwYD
# VQQDExBJbm5lcnZhdGlvbiwgTExDMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIB
# CgKCAQEA24OX3nR2gF8WQ+Te3uJSrV3kOUoxUkczclPxyRXzsfxmkOu4sGU1EeCk
# 5R1fBk68dXuvFbdPUJcKTF2l8kFj6F5a6Ujr5OB9NCBurl96Y+F99RTmYMH0IClt
# YRuWfN1TqNgn9Cn0+uoVwQs8VLFV27y3nIu/ry+eHF5UPvyAnYxFnKkehV4xmibr
# gezzm23MVar0F4p95O8VoJypfvY+gWiPutkrtyRT/cIAcS+7F0F5ZGqEC6H1N8DY
# dXW0NiBF8F5/lWp9K2IGAzVv/vlEXBGTtM9LP5FrEGXr6CeJrmuPfom73ugqe851
# YFxeewyDfVs7fu5tRKsDz0N1EH2FdwIDAQABo4IBxTCCAcEwHwYDVR0jBBgwFoAU
# WsS5eyoKo6XqcQPAYPkt9mV1DlgwHQYDVR0OBBYEFFRsPiAppJf9S2b8T7Edzfbz
# r5EFMA4GA1UdDwEB/wQEAwIHgDATBgNVHSUEDDAKBggrBgEFBQcDAzB3BgNVHR8E
# cDBuMDWgM6Axhi9odHRwOi8vY3JsMy5kaWdpY2VydC5jb20vc2hhMi1hc3N1cmVk
# LWNzLWcxLmNybDA1oDOgMYYvaHR0cDovL2NybDQuZGlnaWNlcnQuY29tL3NoYTIt
# YXNzdXJlZC1jcy1nMS5jcmwwTAYDVR0gBEUwQzA3BglghkgBhv1sAwEwKjAoBggr
# BgEFBQcCARYcaHR0cHM6Ly93d3cuZGlnaWNlcnQuY29tL0NQUzAIBgZngQwBBAEw
# gYQGCCsGAQUFBwEBBHgwdjAkBggrBgEFBQcwAYYYaHR0cDovL29jc3AuZGlnaWNl
# cnQuY29tME4GCCsGAQUFBzAChkJodHRwOi8vY2FjZXJ0cy5kaWdpY2VydC5jb20v
# RGlnaUNlcnRTSEEyQXNzdXJlZElEQ29kZVNpZ25pbmdDQS5jcnQwDAYDVR0TAQH/
# BAIwADANBgkqhkiG9w0BAQsFAAOCAQEA42BdCcQLqAX+7j/dnGh6POSTdLcJHx8E
# 9LG0XIjHpf34nevqvzvoujrSWAfWd3++NpR51qaPDvBw6zZvi8+rf9AHXpHt7hfn
# 25eswIhJ0ipQKdpO3lqTPqcLVovovkqQFTNdkaYpPJLkE/HdVBzC8c76DOhHVyGE
# pMGztfsrUhsZj0mr7IGJVnquwYCHJCuVx7zJ818B3fDccdCwmkl4ZKWGpdgWpnqP
# awMCZtGMbZCrzMfxiJvF0fyvUL1G+f2QvhX71TXrUAeZnC9ETOh+G1F2PvfuXXvx
# h8kes9vFEXvpfdAAvW2fK+GFxYP84pcZ+N+Ug4MZM66kef+/Er6S9zCCBTAwggQY
# oAMCAQICEAQJGBtf1btmdVNDtW+VUAgwDQYJKoZIhvcNAQELBQAwZTELMAkGA1UE
# BhMCVVMxFTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2lj
# ZXJ0LmNvbTEkMCIGA1UEAxMbRGlnaUNlcnQgQXNzdXJlZCBJRCBSb290IENBMB4X
# DTEzMTAyMjEyMDAwMFoXDTI4MTAyMjEyMDAwMFowcjELMAkGA1UEBhMCVVMxFTAT
# BgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEx
# MC8GA1UEAxMoRGlnaUNlcnQgU0hBMiBBc3N1cmVkIElEIENvZGUgU2lnbmluZyBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAPjTsxx/DhGvZ3cH0wsx
# SRnP0PtFmbE620T1f+Wondsy13Hqdp0FLreP+pJDwKX5idQ3Gde2qvCchqXYJawO
# eSg6funRZ9PG+yknx9N7I5TkkSOWkHeC+aGEI2YSVDNQdLEoJrskacLCUvIUZ4qJ
# RdQtoaPpiCwgla4cSocI3wz14k1gGL6qxLKucDFmM3E+rHCiq85/6XzLkqHlOzEc
# z+ryCuRXu0q16XTmK/5sy350OTYNkO/ktU6kqepqCquE86xnTrXE94zRICUj6whk
# PlKWwfIPEvTFjg/BougsUfdzvL2FsWKDc0GCB+Q4i2pzINAPZHM8np+mM6n9Gd8l
# k9ECAwEAAaOCAc0wggHJMBIGA1UdEwEB/wQIMAYBAf8CAQAwDgYDVR0PAQH/BAQD
# AgGGMBMGA1UdJQQMMAoGCCsGAQUFBwMDMHkGCCsGAQUFBwEBBG0wazAkBggrBgEF
# BQcwAYYYaHR0cDovL29jc3AuZGlnaWNlcnQuY29tMEMGCCsGAQUFBzAChjdodHRw
# Oi8vY2FjZXJ0cy5kaWdpY2VydC5jb20vRGlnaUNlcnRBc3N1cmVkSURSb290Q0Eu
# Y3J0MIGBBgNVHR8EejB4MDqgOKA2hjRodHRwOi8vY3JsNC5kaWdpY2VydC5jb20v
# RGlnaUNlcnRBc3N1cmVkSURSb290Q0EuY3JsMDqgOKA2hjRodHRwOi8vY3JsMy5k
# aWdpY2VydC5jb20vRGlnaUNlcnRBc3N1cmVkSURSb290Q0EuY3JsME8GA1UdIARI
# MEYwOAYKYIZIAYb9bAACBDAqMCgGCCsGAQUFBwIBFhxodHRwczovL3d3dy5kaWdp
# Y2VydC5jb20vQ1BTMAoGCGCGSAGG/WwDMB0GA1UdDgQWBBRaxLl7KgqjpepxA8Bg
# +S32ZXUOWDAfBgNVHSMEGDAWgBRF66Kv9JLLgjEtUYunpyGd823IDzANBgkqhkiG
# 9w0BAQsFAAOCAQEAPuwNWiSz8yLRFcgsfCUpdqgdXRwtOhrE7zBh134LYP3DPQ/E
# r4v97yrfIFU3sOH20ZJ1D1G0bqWOWuJeJIFOEKTuP3GOYw4TS63XX0R58zYUBor3
# nEZOXP+QsRsHDpEV+7qvtVHCjSSuJMbHJyqhKSgaOnEoAjwukaPAJRHinBRHoXpo
# aK+bp1wgXNlxsQyPu6j4xRJon89Ay0BEpRPw5mQMJQhCMrI2iiQC/i9yfhzXSUWW
# 6Fkd6fp0ZGuy62ZD2rOwjNXpDd32ASDOmTFjPQgaGLOBm0/GkxAG/AeB+ova+YJJ
# 92JuoVP6EpQYhS6SkepobEQysmah5xikmmRR7zCCBmowggVSoAMCAQICEAMBmgI6
# /1ixa9bV6uYX8GYwDQYJKoZIhvcNAQEFBQAwYjELMAkGA1UEBhMCVVMxFTATBgNV
# BAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEhMB8G
# A1UEAxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xMB4XDTE0MTAyMjAwMDAwMFoX
# DTI0MTAyMjAwMDAwMFowRzELMAkGA1UEBhMCVVMxETAPBgNVBAoTCERpZ2lDZXJ0
# MSUwIwYDVQQDExxEaWdpQ2VydCBUaW1lc3RhbXAgUmVzcG9uZGVyMIIBIjANBgkq
# hkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAo2Rd/Hyz4II14OD2xirmSXU7zG7gU6mf
# H2RZ5nxrf2uMnVX4kuOe1VpjWwJJUNmDzm9m7t3LhelfpfnUh3SIRDsZyeX1kZ/G
# FDmsJOqoSyyRicxeKPRktlC39RKzc5YKZ6O+YZ+u8/0SeHUOplsU/UUjjoZEVX0Y
# hgWMVYd5SEb3yg6Np95OX+Koti1ZAmGIYXIYaLm4fO7m5zQvMXeBMB+7NgGN7yfj
# 95rwTDFkjePr+hmHqH7P7IwMNlt6wXq4eMfJBi5GEMiN6ARg27xzdPpO2P6qQPGy
# znBGg+naQKFZOtkVCVeZVjCT88lhzNAIzGvsYkKRrALA76TwiRGPdwIDAQABo4ID
# NTCCAzEwDgYDVR0PAQH/BAQDAgeAMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAww
# CgYIKwYBBQUHAwgwggG/BgNVHSAEggG2MIIBsjCCAaEGCWCGSAGG/WwHATCCAZIw
# KAYIKwYBBQUHAgEWHGh0dHBzOi8vd3d3LmRpZ2ljZXJ0LmNvbS9DUFMwggFkBggr
# BgEFBQcCAjCCAVYeggFSAEEAbgB5ACAAdQBzAGUAIABvAGYAIAB0AGgAaQBzACAA
# QwBlAHIAdABpAGYAaQBjAGEAdABlACAAYwBvAG4AcwB0AGkAdAB1AHQAZQBzACAA
# YQBjAGMAZQBwAHQAYQBuAGMAZQAgAG8AZgAgAHQAaABlACAARABpAGcAaQBDAGUA
# cgB0ACAAQwBQAC8AQwBQAFMAIABhAG4AZAAgAHQAaABlACAAUgBlAGwAeQBpAG4A
# ZwAgAFAAYQByAHQAeQAgAEEAZwByAGUAZQBtAGUAbgB0ACAAdwBoAGkAYwBoACAA
# bABpAG0AaQB0ACAAbABpAGEAYgBpAGwAaQB0AHkAIABhAG4AZAAgAGEAcgBlACAA
# aQBuAGMAbwByAHAAbwByAGEAdABlAGQAIABoAGUAcgBlAGkAbgAgAGIAeQAgAHIA
# ZQBmAGUAcgBlAG4AYwBlAC4wCwYJYIZIAYb9bAMVMB8GA1UdIwQYMBaAFBUAEisT
# mLKZB+0e36K+Vw0rZwLNMB0GA1UdDgQWBBRhWk0ktkkynUoqeRqDS/QeicHKfTB9
# BgNVHR8EdjB0MDigNqA0hjJodHRwOi8vY3JsMy5kaWdpY2VydC5jb20vRGlnaUNl
# cnRBc3N1cmVkSURDQS0xLmNybDA4oDagNIYyaHR0cDovL2NybDQuZGlnaWNlcnQu
# Y29tL0RpZ2lDZXJ0QXNzdXJlZElEQ0EtMS5jcmwwdwYIKwYBBQUHAQEEazBpMCQG
# CCsGAQUFBzABhhhodHRwOi8vb2NzcC5kaWdpY2VydC5jb20wQQYIKwYBBQUHMAKG
# NWh0dHA6Ly9jYWNlcnRzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRENB
# LTEuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCdJX4bM02yJoFcm4bOIyAPgIfliP//
# sdRqLDHtOhcZcRfNqRu8WhY5AJ3jbITkWkD73gYBjDf6m7GdJH7+IKRXrVu3mrBg
# JuppVyFdNC8fcbCDlBkFazWQEKB7l8f2P+fiEUGmvWLZ8Cc9OB0obzpSCfDscGLT
# Ykuw4HOmksDTjjHYL+NtFxMG7uQDthSr849Dp3GdId0UyhVdkkHa+Q+B0Zl0DSbE
# Dn8btfWg8cZ3BigV6diT5VUW8LsKqxzbXEgnZsijiwoc5ZXarsQuWaBh3drzbaJh
# 6YoLbewSGL33VVRAA5Ira8JRwgpIr7DUbuD0FAo6G+OPPcqvao173NhEMIIGzTCC
# BbWgAwIBAgIQBv35A5YDreoACus/J7u6GzANBgkqhkiG9w0BAQUFADBlMQswCQYD
# VQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGln
# aWNlcnQuY29tMSQwIgYDVQQDExtEaWdpQ2VydCBBc3N1cmVkIElEIFJvb3QgQ0Ew
# HhcNMDYxMTEwMDAwMDAwWhcNMjExMTEwMDAwMDAwWjBiMQswCQYDVQQGEwJVUzEV
# MBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29t
# MSEwHwYDVQQDExhEaWdpQ2VydCBBc3N1cmVkIElEIENBLTEwggEiMA0GCSqGSIb3
# DQEBAQUAA4IBDwAwggEKAoIBAQDogi2Z+crCQpWlgHNAcNKeVlRcqcTSQQaPyTP8
# TUWRXIGf7Syc+BZZ3561JBXCmLm0d0ncicQK2q/LXmvtrbBxMevPOkAMRk2T7It6
# NggDqww0/hhJgv7HxzFIgHweog+SDlDJxofrNj/YMMP/pvf7os1vcyP+rFYFkPAy
# IRaJxnCI+QWXfaPHQ90C6Ds97bFBo+0/vtuVSMTuHrPyvAwrmdDGXRJCgeGDboJz
# PyZLFJCuWWYKxI2+0s4Grq2Eb0iEm09AufFM8q+Y+/bOQF1c9qjxL6/siSLyaxhl
# scFzrdfx2M8eCnRcQrhofrfVdwonVnwPYqQ/MhRglf0HBKIJAgMBAAGjggN6MIID
# djAOBgNVHQ8BAf8EBAMCAYYwOwYDVR0lBDQwMgYIKwYBBQUHAwEGCCsGAQUFBwMC
# BggrBgEFBQcDAwYIKwYBBQUHAwQGCCsGAQUFBwMIMIIB0gYDVR0gBIIByTCCAcUw
# ggG0BgpghkgBhv1sAAEEMIIBpDA6BggrBgEFBQcCARYuaHR0cDovL3d3dy5kaWdp
# Y2VydC5jb20vc3NsLWNwcy1yZXBvc2l0b3J5Lmh0bTCCAWQGCCsGAQUFBwICMIIB
# Vh6CAVIAQQBuAHkAIAB1AHMAZQAgAG8AZgAgAHQAaABpAHMAIABDAGUAcgB0AGkA
# ZgBpAGMAYQB0AGUAIABjAG8AbgBzAHQAaQB0AHUAdABlAHMAIABhAGMAYwBlAHAA
# dABhAG4AYwBlACAAbwBmACAAdABoAGUAIABEAGkAZwBpAEMAZQByAHQAIABDAFAA
# LwBDAFAAUwAgAGEAbgBkACAAdABoAGUAIABSAGUAbAB5AGkAbgBnACAAUABhAHIA
# dAB5ACAAQQBnAHIAZQBlAG0AZQBuAHQAIAB3AGgAaQBjAGgAIABsAGkAbQBpAHQA
# IABsAGkAYQBiAGkAbABpAHQAeQAgAGEAbgBkACAAYQByAGUAIABpAG4AYwBvAHIA
# cABvAHIAYQB0AGUAZAAgAGgAZQByAGUAaQBuACAAYgB5ACAAcgBlAGYAZQByAGUA
# bgBjAGUALjALBglghkgBhv1sAxUwEgYDVR0TAQH/BAgwBgEB/wIBADB5BggrBgEF
# BQcBAQRtMGswJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmRpZ2ljZXJ0LmNvbTBD
# BggrBgEFBQcwAoY3aHR0cDovL2NhY2VydHMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0
# QXNzdXJlZElEUm9vdENBLmNydDCBgQYDVR0fBHoweDA6oDigNoY0aHR0cDovL2Ny
# bDMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEUm9vdENBLmNybDA6oDig
# NoY0aHR0cDovL2NybDQuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEUm9v
# dENBLmNybDAdBgNVHQ4EFgQUFQASKxOYspkH7R7for5XDStnAs0wHwYDVR0jBBgw
# FoAUReuir/SSy4IxLVGLp6chnfNtyA8wDQYJKoZIhvcNAQEFBQADggEBAEZQPsm3
# KCSnOB22WymvUs9S6TFHq1Zce9UNC0Gz7+x1H3Q48rJcYaKclcNQ5IK5I9G6OoZy
# rTh4rHVdFxc0ckeFlFbR67s2hHfMJKXzBBlVqefj56tizfuLLZDCwNK1lL1eT7EF
# 0g49GqkUW6aGMWKoqDPkmzmnxPXOHXh2lCVz5Cqrz5x2S+1fwksW5EtwTACJHvzF
# ebxMElf+X+EevAJdqP77BzhPDcZdkbkPZ0XN1oPt55INjbFpjE/7WeAjD9KqrgB8
# 7pxCDs+R1ye3Fu4Pw718CqDuLAhVhSK46xgaTfwqIa1JMYNHlXdx3LEbS0scEJx3
# FMGdTy9alQgpECYxggQ7MIIENwIBATCBhjByMQswCQYDVQQGEwJVUzEVMBMGA1UE
# ChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMTEwLwYD
# VQQDEyhEaWdpQ2VydCBTSEEyIEFzc3VyZWQgSUQgQ29kZSBTaWduaW5nIENBAhAB
# bOl77YhPJjl7V/MavkwsMAkGBSsOAwIaBQCgeDAYBgorBgEEAYI3AgEMMQowCKAC
# gAChAoAAMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQBgjcCAQsx
# DjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBTwZIhV9qmBEBawkr/gMoEF
# W3UN6jANBgkqhkiG9w0BAQEFAASCAQAhJNKwQqH2kCL0owSAR05rJiOYnpBt7wfl
# 6/4P5cohuHxBLuNA+CD4YIyjWck/EkhYwLG/+mAR1hgKn9sIZ0u9bg80xrkm0ACE
# lGvGnC1Y+VJPaoC3c6Ur6+cebxwJop1cmwxPSpwyzwb+EFHQHWlkFdJCEPIKJuHo
# zn79r5rLoXFSrMiHeVw4f55CV0kB/az/fQ9p0eDHUI78296UIz5XtJqeBL6DMBOk
# XDx7ENTZtIOi+V5AGK/RZ8eoYxKPjjClNPDfVusg7bjEaC1VnTO5x/C4m6xUrSv9
# 39cXOMuSGhBZ36sLMwpaHeGB6YyCzU1pHrMGo8W40Sqh1a70knssoYICDzCCAgsG
# CSqGSIb3DQEJBjGCAfwwggH4AgEBMHYwYjELMAkGA1UEBhMCVVMxFTATBgNVBAoT
# DERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEhMB8GA1UE
# AxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xAhADAZoCOv9YsWvW1ermF/BmMAkG
# BSsOAwIaBQCgXTAYBgkqhkiG9w0BCQMxCwYJKoZIhvcNAQcBMBwGCSqGSIb3DQEJ
# BTEPFw0xODAzMjgwMjA5MThaMCMGCSqGSIb3DQEJBDEWBBR30IDxaeRAQBsR+NLW
# HReeIYmjSDANBgkqhkiG9w0BAQEFAASCAQAbDHEzFwVfbmhEdrFDZdIemFZnxiC+
# osXsIapSpeaGz/q4p3jwycRkDgL5uySq9dB95k2CP+Ymbx9gGCP3Jpq9luvBbneD
# LBxsZKzhEBY67la1tTEIMrnC3pNB4xuahKHJNfkGkSxMyJAv3bI+lTTSSQ+eE3bi
# zGs4wSIeoghQ1z0/7umMZDK67Ghtzi/uQouXUrNDiT7u82nFqEg6VgS2T4/zsHbn
# W7B5VvovVA4HS1j9EvxCDiEd/VqtloohBS+x+OTEFzUNW893hKqszRhO24Sb3KDt
# cF0byUqVGWHros1nLWyEesnUfMoYlENTRaCI5y7MnqFwrHRmruxUXCea
# SIG # End signature block
